/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "about.h"
#include "main.h"
#include "runEmbossAppLocal.h"
#include "runEmbossAppOnServer.h"

#include <QProcess>
#include <QDesktopServices>


aboutForm::aboutForm(QWidget *parent, kembossSettings *fksettings) : QDialog(parent)
{
   setupUi(this);
   setAttribute(Qt::WA_DeleteOnClose);

   ksettings = fksettings;

   // Get Qt version
   QtVersionLabel->setText("Qt version "+QString(qVersion()));

   // Get EMBOSS version
   if (ksettings->remoteEmboss()) {
       runEmbossVersion = new krunEmbossAppOnServer(this, ksettings);
   } else {
       runEmbossVersion = new krunEmbossAppLocal(this, ksettings);
   }
   runEmbossVersion->setShowProgressWindow(false);
   runEmbossVersion->setEmbossApp("embossversion");
   connect(runEmbossVersion, SIGNAL(applicationExecuted(int)), this, SLOT(showEmbossVersion()));
   runEmbossVersion->runApp();

   // Check for kemboss update
   QUrl versionUrl("http://kemboss.org/latest_kemboss_version");
   m_pVersionCtrl = new versionCheck(versionUrl, this);

   connect(m_pVersionCtrl, SIGNAL(downloaded()), SLOT(getLatestKembossVersion()));
}

aboutForm::~aboutForm()
{
    runEmbossVersion->~krunEmbossApp();
    if (runEmbossUpdate != nullptr)
        runEmbossUpdate->~krunEmbossApp();
}

void aboutForm::showEmbossVersion()
{
    QRegExp rx("(\\d)\\.(\\d)\\.(\\d)");
    int pos = rx.indexIn(runEmbossVersion->programOutput());
    if (pos > -1) {
        EmbossVersionLabel->setText("EMBOSS version "+rx.cap(0));
        int majorVer = rx.cap(1).toInt();
        int minorVer = rx.cap(2).toInt();
        if (majorVer >= 6 && minorVer >= 5) {
            // Check for EMBOSS update
            if (ksettings->remoteEmboss()) {
                runEmbossUpdate = new krunEmbossAppOnServer(this, ksettings);
            } else {
                runEmbossUpdate = new krunEmbossAppLocal(this, ksettings);
            }
            runEmbossUpdate->setShowProgressWindow(false);
            runEmbossUpdate->setEmbossApp("embossupdate");
            connect(runEmbossUpdate, SIGNAL(applicationExecuted(int)), this, SLOT(showEmbossUpdate()));
            runEmbossUpdate->runApp();
        } else {
            EmbossVersionLabel->setText(EmbossVersionLabel->text()+" (Update available)");
        }
    }
}

void aboutForm::showEmbossUpdate()
{
    if (runEmbossUpdate->programOutput().trimmed().contains("is the latest version", Qt::CaseInsensitive))
        EmbossVersionLabel->setText(EmbossVersionLabel->text()+" (Latest version)");
    else
        EmbossVersionLabel->setText(EmbossVersionLabel->text()+" (Update available)");
}

void aboutForm::getLatestKembossVersion()
{
    QString updateVersion = m_pVersionCtrl->downloadedData();
    if (updateVersion == "")
        KembossVersionLabel->setText("kemboss version "+QCoreApplication::applicationVersion());
    else {
        if (updateVersion == QCoreApplication::applicationVersion())
            KembossVersionLabel->setText("kemboss version "+QCoreApplication::applicationVersion()+" (Latest version)");
        else
            KembossVersionLabel->setText("kemboss version "+QCoreApplication::applicationVersion()+" (Updated version "+updateVersion+" available)");
    }
}

void aboutForm::on_GPLButton_clicked()
{
   QString filename;
   filename = ksettings->kembossData()+"../copying_GPL_V2";
   #ifdef Q_OS_WIN32
   filename += ".txt";
   #endif
   #ifdef Q_OS_MAC
   filename += ".html";
   #endif
   if (QFile::exists(filename))
      QDesktopServices::openUrl(QUrl::fromLocalFile(filename));
}

void aboutForm::on_CopyrightButton_clicked()
{
   QString filename;
   filename = ksettings->kembossData()+"../copyright";
   #ifdef Q_OS_WIN32
   filename += ".txt";
   #endif
   #ifdef Q_OS_MAC
   filename += ".html";
   #endif
   if (QFile::exists(filename))
      QDesktopServices::openUrl(QUrl::fromLocalFile(filename));
}

void aboutForm::on_URLButton_clicked()
{
   QDesktopServices::openUrl(QUrl("http://www.kemboss.org/"));
}



versionCheck::versionCheck(QUrl versionUrl, QObject *parent) : QObject(parent)
{
    connect(&m_WebCtrl, SIGNAL(finished(QNetworkReply*)),
                SLOT(versionDownloaded(QNetworkReply*)));
    QNetworkRequest request(versionUrl);
    m_WebCtrl.get(request);
}

versionCheck::~versionCheck()
{

}

void versionCheck::versionDownloaded(QNetworkReply* pReply)
{
    m_DownloadedData = pReply->readAll();
    //emit a signal
    emit downloaded();
}

QByteArray versionCheck::downloadedData() const
{
    return m_DownloadedData;
}
