/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2017 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include <QtCore/QCoreApplication>
#include <QtNetwork>
#include <QSettings>
#include <QDir>
#include <QCommandLineParser>

#include "kserver.h"
#include "sql.h"
#include "../main.h"

int main(int argc, char *argv[])
{
    QCoreApplication app(argc, argv);

    QCoreApplication::setOrganizationName("kemboss.org");
    QCoreApplication::setOrganizationDomain("kemboss.org");
    QCoreApplication::setApplicationName("kembossd");
    QCoreApplication::setApplicationVersion(QString(KEMBOSS_SERVER_MAJOR_VERSION)+"."+QString(KEMBOSS_SERVER_MINOR_VERSION)+"."+QString(KEMBOSS_SERVER_MINIMUM_VERSION));

    QCommandLineParser parser;
    parser.setApplicationDescription("Server to enable access to the sequence analysis package EMBOSS via the kemboss graphical client.");
    parser.addHelpOption();
    parser.addVersionOption();

    // Process the actual command line arguments given by the user
    parser.process(app);

    QTextStream qstdout(stdout);
    QTextStream qstderr(stderr);

    // Search the server configuration file
    QString configFile = kembossServer::getServerConfigFile();

    // Check for server configuration file
    qstdout << "Reading server configuration in " << configFile << endl;
    if (!QFile::exists(configFile)) {
        qstderr << "kemboss server: Could not find server configuration. Run kembossd-admin before starting the server." << endl;
        exit(1);
    }

    qstdout << "Starting kemboss server..." << endl;
    KembossServer *server;
    server = new KembossServer();

    // Configure server
    QSettings settings(configFile, QSettings::IniFormat);
    server->setWelcomeMessage(settings.value("WelcomeMessage", "Welcome on kemboss server").toString());
    server->setAcceptAllUsers(settings.value("AcceptAllNewUsers", false).toBool());
    if (!server->setEmbossFileDir(settings.value("UserDir", "").toString()))
        exit(1);
    server->setEmbossBinDir(settings.value("EmbossBinDir", "").toString());
    if (!server->loadEmbossProgramList(settings.value("EmbossProgramList", "").toString())) {
        qstderr << "Could not open file containing the allowed EMBOSS programs. Exiting kembossd.";
        exit(1);
    }
    server->setMaxEmbossProcessTime(settings.value("MaxEmbossProgramExecTime", 3600).toUInt());
    server->setMaxThreadNumber(settings.value("MaxNumberEmbossPrograms", 10).toUInt());
    server->setLogfile(settings.value("Logfile", "stdout").toString());

    // Check if SSL is supported by the installed Qt libraries
    if (!QSslSocket::supportsSsl()) {
        qstdout << "The installed Qt libraries do not support SSL connections. Terminating..." << endl;
        exit(2);
    } else
        qstdout << "Using SSL library version: " << QSslSocket::sslLibraryVersionString() << endl;

    settings.beginGroup("SSL_connection");
    server->setPrivateKeyFile(settings.value("Privkey", "./privkey.pem").toString());
    server->setCACertsPath(settings.value("Cacert", "./cacert.pem").toString());
    server->setLocalCertFile(settings.value("Localcert", "./localcert.pem").toString());
    settings.endGroup();

    settings.beginGroup("EMBASSY_packages");
    server->setPhylipnewInstalled(settings.value("PHYLIPNEW", false).toBool());
    settings.endGroup();

    if (!server->listen(QHostAddress::Any, settings.value("Port", 5757).toUInt())) {
        qstderr << "kemboss server: Unable to start the server: " << server->errorString() << endl;
        exit(1);
    }
    qstdout << "The server is running on Port: " << server->serverPort() << endl;

    // Check Database
    if (!ksql::checkDb()) {
        qstderr << "kemboss server: Unable to connect to SQL database." << endl;
        exit(1);
    }

    qstdout << "Server commands:" << endl;
    qstdout << "l: List running EMBOSS programs" << endl;
    qstdout << "t: Terminate running EMBOSS program" << endl;
    qstdout << "v: Show kembossd version" << endl;
    qstdout << "x: Shutdown server" << endl << endl;

    qstdout << "Waiting for client connections..." << endl << endl;

    qstdout << "Enter server command: " << flush;

    server->setServerStatus(kemboss::Ready);

    return app.exec();
}
