/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "programWindowClass.h"
#include "kemboss.h"
#include "main.h"
#include "configDlg.h"
#include "associated_qualifiers.h"
#include "finddialog.h"
#include "findreplacedialog.h"
#include "loadFileThread.h"
#include "runEmbossAppLocal.h"
#include "runEmbossAppOnServer.h"

#include <cstdio>
#include <QMessageBox>
#include <QFileDialog>
#include <QClipboard>
#include <QCloseEvent>
#include <QWhatsThis>
#include <QtPrintSupport/QPrintDialog>
#include <QtPrintSupport/QPageSetupDialog>

kprogramWindow::kprogramWindow(kembossWindow *parent) : kembossWindow(parent)
{
   setAttribute(Qt::WA_DeleteOnClose);
   kembossParent = parent;
   ksettings = &qobject_cast<kemboss *>(kembossParent)->ksettings;

   kpwInputTextEdit = nullptr;
   kpwInputFileLineEdit = nullptr;
   kpwInputTextEdit2 = nullptr;
   kpwInputFileLineEdit2 = nullptr;
   kpwOutputFileLineEdit = nullptr;
   kpwOutputTextEdit = nullptr;
   currentTextEdit = nullptr;

   seqFormatInput = 0;
   seqFormatOutput = 0;

   highlighter = new Highlighter(nullptr, ksettings);

   if (ksettings->remoteEmboss()) {
       runEmbossApp = new krunEmbossAppOnServer(this, ksettings);
   } else {
       runEmbossApp = new krunEmbossAppLocal(this, ksettings);
   }
   connect(runEmbossApp, SIGNAL(applicationExecuted(int)), this, SLOT(embossExecutionFinished(int)));

   // connect(&loadThread, SIGNAL(loadedFile(QString, bool)), this, SLOT(appendText(QString, bool)));

   m_findDialog = new FindDialog(this);
   m_findDialog->setModal(false);
   m_findReplaceDialog = new FindReplaceDialog(this);
   m_findReplaceDialog->setModal(false);

   loadPrinterSettings();
}

void kprogramWindow::postWindowCreated()
{
   createActions();
   createMenus();
   createToolbars();
   findControls();
   readSettings();
   QSettings *settings;
   if (ksettings->portableApp()) {
       settings = new QSettings(ksettings->kembossPortableConfigDir()+objectName(), QSettings::IniFormat);
   } else {
       settings = new QSettings(QSettings::UserScope, "kemboss.org", objectName());
   }
   m_findDialog->readSettings(settings);
   m_findReplaceDialog->readSettings(settings);
   tabChanged(0);
   setTextEditFonts();
   createRecentFiles();

   tempDir = new QTemporaryDir(QDir::tempPath()+"/kemboss");

   runEmbossApp->setEmbossApp(objectName().remove(0,1)); // The EMBOSS application could be changed in beforeRunningProgram() to allow single GUI windows to start different programs
   if (ksettings->openLastSettings()) {
      QString fileName = settings->value("last_settings").toString();
      if (QFile::exists(fileName) && loadSettings(fileName)) {
         setCurrentFile(fileName);
         return;
      }
   }
   newSettings();
}

void kprogramWindow::createActions()
{
   newAct = new QAction(QIcon(":/new/prefix1/filenew.png"), tr("&New"), this);
   newAct->setShortcut(tr("Ctrl+N"));
   newAct->setStatusTip(tr("Set to default settings"));
   connect(newAct, SIGNAL(triggered()), this, SLOT(newSettings()));

   openAct = new QAction(QIcon(":/new/prefix1/fileopen.png"), tr("&Open..."), this);
   openAct->setShortcut(tr("Ctrl+O"));
   openAct->setStatusTip(tr("Open settings from file"));
   connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

   saveAct = new QAction(QIcon(":/new/prefix1/filesave.png"), tr("&Save"), this);
   saveAct->setShortcut(tr("Ctrl+S"));
   saveAct->setStatusTip(tr("Save settings to current file"));
   connect(saveAct, SIGNAL(triggered()), this, SLOT(save()));
   saveAct->setEnabled(false);

   saveAsAct = new QAction(QIcon(":/new/prefix1/filesaveas.png"), tr("Save &As..."), this);
   saveAsAct->setStatusTip(tr("Save settings to new file"));
   connect(saveAsAct, SIGNAL(triggered()), this, SLOT(saveAs()));

   exitAct = new QAction(QIcon(":/new/prefix1/exit.png"), tr("&Quit"), this);
   exitAct->setShortcut(tr("Ctrl+Q"));
   exitAct->setStatusTip(tr("Close window"));
   connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));

   propertiesAct = new QAction(QIcon(":/new/prefix1/document-properties.png"), tr("Print properties"), this);
   propertiesAct->setStatusTip(tr("Print properties"));
   connect(propertiesAct, SIGNAL(triggered()), this, SLOT(documentProperties()));

   printAct = new QAction(QIcon(":/new/prefix1/document-print.png"), tr("Print"), this);
   printAct->setShortcut(tr("Ctrl+P"));
   printAct->setStatusTip(tr("Print"));
   connect(printAct, SIGNAL(triggered()), this, SLOT(printDocument()));

   runAct = new QAction(QIcon(":/new/prefix1/run.png"), tr("&Run"), this);
   runAct->setShortcut(tr("Ctrl+R"));
   runAct->setStatusTip(tr("Execute EMBOSS application"));
   connect(runAct, SIGNAL(triggered()), this, SLOT(runEmboss()));

   addParameterAct = new QAction(QIcon(":/new/prefix1/add.png"), tr("Add Parameter"), this);
   addParameterAct->setStatusTip(tr("Add configuration parameter for EMBOSS application"));
   connect(addParameterAct, SIGNAL(triggered()), this, SLOT(addParameter()));

   copyCommandlineAct = new QAction(QIcon(":/new/prefix1/document-export.png"), tr("Export commandline"), this);
   copyCommandlineAct->setStatusTip(tr("Export commandline for EMBOSS application to clipboard"));
   connect(copyCommandlineAct, SIGNAL(triggered()), this, SLOT(copyCommandline()));

   cutAct = new QAction(QIcon(":/new/prefix1/cut.png"), tr("Cu&t"), this);
   cutAct->setShortcut(tr("Ctrl+X"));
   cutAct->setStatusTip(tr("Cut the current selection's contents to the clipboard"));
   cutAct->setEnabled(false);

   copyAct = new QAction(QIcon(":/new/prefix1/copy.png"), tr("&Copy"), this);
   copyAct->setShortcut(tr("Ctrl+C"));
   copyAct->setStatusTip(tr("Copy the current selection's contents to the clipboard"));
   copyAct->setEnabled(false);

   pasteAct = new QAction(QIcon(":/new/prefix1/paste.png"), tr("&Paste"), this);
   pasteAct->setShortcut(tr("Ctrl+V"));
   pasteAct->setStatusTip(tr("Paste the clipboard's contents into the current selection"));

   undoAct = new QAction(QIcon(":/new/prefix1/undo.png"), tr("&Undo"), this);
   undoAct->setShortcut(tr("Ctrl+Z"));
   undoAct->setStatusTip(tr("Undo last operation"));
   undoAct->setEnabled(false);

   redoAct = new QAction(QIcon(":/new/prefix1/redo.png"), tr("&Redo"), this);
   QList<QKeySequence> redoShortcuts;
   redoShortcuts << tr("Ctrl+Y") << tr("Shift+Ctrl+Z");
   redoAct->setShortcuts(redoShortcuts);
   redoAct->setStatusTip(tr("Redo last operation"));
   redoAct->setEnabled(false);

   findAct = new QAction(QIcon(":/new/prefix1/edit-find.png"), tr("Find..."), this);
   findAct->setShortcut(tr("Ctrl+F"));
   findAct->setStatusTip(tr("Find text or sequence"));
   connect(findAct, SIGNAL(triggered()), this, SLOT(find()));

   findNextAct = new QAction(QIcon(":/new/prefix1/find-next.png"), tr("Find next"), this);
   findNextAct->setShortcut(tr("F3"));
   findNextAct->setStatusTip(tr("Find next text or sequence"));
   connect(findNextAct, SIGNAL(triggered()), m_findDialog, SLOT(findNext()));

   findReplaceAct = new QAction(QIcon(":/new/prefix1/edit-find-replace.png"), tr("Find/Replace..."), this);
   findReplaceAct->setShortcut(tr("Ctrl+H"));
   findReplaceAct->setStatusTip(tr("Find and replace text or sequence"));
   connect(findReplaceAct, SIGNAL(triggered()), this, SLOT(findReplace()));

   rotateLeftAct = new QAction(QIcon(":/new/prefix1/object-rotate-left.png"), tr("Rotate left"), this);
   rotateLeftAct->setStatusTip(tr("Rotate graph left"));
   connect(rotateLeftAct, SIGNAL(triggered()), this, SLOT(rotateLeft()));

   rotateRightAct = new QAction(QIcon(":/new/prefix1/object-rotate-right.png"), tr("Rotate right"), this);
   rotateRightAct->setStatusTip(tr("Rotate graph right"));
   connect(rotateRightAct, SIGNAL(triggered()), this, SLOT(rotateRight()));

   zoomInAct = new QAction(QIcon(":/new/prefix1/zoom-in.png"), tr("Zoom in"), this);
   zoomInAct->setStatusTip(tr("Zoom in graph"));
   connect(zoomInAct, SIGNAL(triggered()), this, SLOT(zoomIn()));

   zoomOutAct = new QAction(QIcon(":/new/prefix1/zoom-out.png"), tr("Zoom out"), this);
   zoomOutAct->setStatusTip(tr("Zoom out graph"));
   connect(zoomOutAct, SIGNAL(triggered()), this, SLOT(zoomOut()));

   zoomOrigAct = new QAction(QIcon(":/new/prefix1/zoom-original.png"), tr("Zoom original"), this);
   zoomOrigAct->setStatusTip(tr("Show graph using original size"));
   connect(zoomOrigAct, SIGNAL(triggered()), this, SLOT(zoomOriginal()));

   highlightSeqAct = new QAction(QIcon(":/new/prefix1/highlight.png"), tr("Highlight sequence"), this);
   highlightSeqAct->setCheckable(true);
   highlightSeqAct->setStatusTip((tr("Highlight sequence using the selected sequence format")));
   connect(highlightSeqAct, SIGNAL(toggled(bool)), this, SLOT(highlightSeq(bool)));

   seqFormatNoneAct = new QAction(tr("None"), this);
   seqFormatNoneAct->setStatusTip(tr("No sequence format selected"));
   connect(seqFormatNoneAct, SIGNAL(triggered()), this, SLOT(seqFormatNone()));

   seqFormatFastaAAAct = new QAction(tr("FASTA protein"), this);
   seqFormatFastaAAAct->setStatusTip(tr("Protein sequence in FASTA format"));
   connect(seqFormatFastaAAAct, SIGNAL(triggered()), this, SLOT(seqFormatFastaAA()));

   seqFormatFastaNucAct = new QAction(tr("FASTA nucleic acid"), this);
   seqFormatFastaNucAct->setStatusTip(tr("Nucleic acid sequence in FASTA format"));
   connect(seqFormatFastaNucAct, SIGNAL(triggered()), this, SLOT(seqFormatFastaNuc()));

   seqFormatAlignmentAAAct = new QAction(tr("Alignment protein"), this);
   seqFormatAlignmentAAAct->setStatusTip(tr("Protein sequence in alignment format"));
   connect(seqFormatAlignmentAAAct, SIGNAL(triggered()), this, SLOT(seqFormatAlignmentAA()));

   seqFormatAlignmentNucAct = new QAction(tr("Alignment nucleic acid"), this);
   seqFormatAlignmentNucAct->setStatusTip(tr("Nucleic acid sequence in alignment format"));
   connect(seqFormatAlignmentNucAct, SIGNAL(triggered()), this, SLOT(seqFormatAlignmentNuc()));

   seqFormatEmblAct = new QAction(tr("EMBL"), this);
   seqFormatEmblAct->setStatusTip(tr("Sequence in EMBL format"));
   connect(seqFormatEmblAct, SIGNAL(triggered()), this, SLOT(seqFormatEmbl()));

   seqFormatGbAAAct = new QAction(tr("Gb Protein"), this);
   seqFormatGbAAAct->setStatusTip(tr("Protein sequence in GB format"));
   connect(seqFormatGbAAAct, SIGNAL(triggered()), this, SLOT(seqFormatGbAA()));

   seqFormatGbNucAct = new QAction(tr("Gb nuleic acid"), this);
   seqFormatGbNucAct->setStatusTip(tr("Nucleic acid sequence in GB format"));
   connect(seqFormatGbNucAct, SIGNAL(triggered()), this, SLOT(seqFormatGbNuc()));

   seqFormatSwissAct = new QAction(tr("Swissprot"), this);
   seqFormatSwissAct->setStatusTip(tr("Sequence in Swissprot format"));
   connect(seqFormatSwissAct, SIGNAL(triggered()), this, SLOT(seqFormatSwiss()));

   helpAct = new QAction(QIcon(":/new/prefix1/help.png"), tr("EMBOSS Help"), this);
   helpAct->setShortcut(tr("Ctrl+?"));
   helpAct->setStatusTip(tr("Show help on EMBOSS program"));
   connect(helpAct, SIGNAL(triggered()), this, SLOT(embossHelp()));

   whatsThisAct = QWhatsThis::createAction(this);

   kembossWindow::createActions();
}

void kprogramWindow::createMenus()
{
   fileMenu = menuBar()->addMenu(tr("&File"));
   fileMenu->addAction(newAct);
   fileMenu->addAction(openAct);

   recentSettingsMenu = fileMenu->addMenu(QIcon(":/new/prefix1/open_recent.png"), tr("Recent Settings"));

   fileMenu->addMenu(recentSettingsMenu);
   fileMenu->addSeparator();
   fileMenu->addAction(saveAct);
   fileMenu->addAction(saveAsAct);
   fileMenu->addSeparator();
   fileMenu->addAction(propertiesAct);
   fileMenu->addAction(printAct);
   fileMenu->addSeparator();
   fileMenu->addAction(exitAct);

   editMenu = menuBar()->addMenu(tr("&Edit"));
   editMenu->addAction(cutAct);
   editMenu->addAction(copyAct);
   editMenu->addAction(pasteAct);
   editMenu->addSeparator();
   editMenu->addAction(undoAct);
   editMenu->addAction(redoAct);
   editMenu->addSeparator();
   editMenu->addAction(findAct);
   editMenu->addAction(findNextAct);
   editMenu->addAction(findReplaceAct);
   editMenu->addSeparator();

   graphMenu = editMenu->addMenu(tr("Graph"));
   graphMenu->addAction(rotateLeftAct);
   graphMenu->addAction(rotateRightAct);
   graphMenu->addSeparator();
   graphMenu->addAction(zoomOutAct);
   graphMenu->addAction(zoomOrigAct);
   graphMenu->addAction(zoomInAct);

   viewMenu = menuBar()->addMenu(tr("&View"));
  /* tabsMenu = viewMenu->addMenu(tr("Ta&bs"));
   QAction *tabAct;
   for (int i = 0; i < kpwMainTab->count(); ++i) {
      tabAct = new QAction(kpwMainTab->tabText(i),this);
      tabAct->setStatusTip(tr("Show ")+kpwMainTab->tabText(i)+tr(" tab"));
      tabAct->setData(i);
      tabAct->setCheckable(true);
      tabAct->setChecked(true);
      connect(tabAct, SIGNAL(triggered()), this, SLOT(viewTab()));
      tabsMenu->addAction(tabAct);
   }*/
   viewMenu->addAction(highlightSeqAct);

   seqFormatMenu = viewMenu->addMenu(tr("Sequence format"));
   seqFormatMenu->addAction(seqFormatNoneAct);
   seqFormatMenu->addAction(seqFormatFastaAAAct);
   seqFormatMenu->addAction(seqFormatFastaNucAct);
   seqFormatMenu->addAction(seqFormatAlignmentAAAct);
   seqFormatMenu->addAction(seqFormatAlignmentNucAct);
   seqFormatMenu->addAction(seqFormatEmblAct);
   seqFormatMenu->addAction(seqFormatGbAAAct);
   seqFormatMenu->addAction(seqFormatGbNucAct);
   seqFormatMenu->addAction(seqFormatSwissAct);

   toolbarsMenu = viewMenu->addMenu(tr("&Toolbars"));
   toolbarsMenu->addAction(viewFileToolbarAct);
   toolbarsMenu->addAction(viewEditToolbarAct);
   toolbarsMenu->addAction(viewViewToolbarAct);
   toolbarsMenu->addAction(viewGraphToolbarAct);
   toolbarsMenu->addAction(viewRunToolbarAct);
   toolbarsMenu->addAction(viewHelpToolbarAct);

   runMenu = menuBar()->addMenu(tr("&Run"));
   runMenu->addAction(addParameterAct);
   runMenu->addAction(copyCommandlineAct);
   runMenu->addAction(runAct);

   menuBar()->addSeparator();
   //This has no effect for most widget styles (e.g., Windows and Mac OS X styles), but for Motif-based styles this makes sure that Help is pushed to the right side of the menu bar.

   helpMenu = menuBar()->addMenu(tr("&Help"));
   helpMenu->addAction(helpAct);
   helpMenu->addAction(whatsThisAct);
   helpMenu->addAction(reportBugAct);
   helpMenu->addAction(aboutAct);
}

void kprogramWindow::createToolbars()
{
   fileToolBar = addToolBar(tr("File"));
   fileToolBar->addAction(exitAct);
   fileToolBar->addAction(newAct);
   fileToolBar->addAction(openAct);
   fileToolBar->addAction(saveAct);
   fileToolBar->addAction(saveAsAct);
   fileToolBar->addSeparator();
   fileToolBar->addAction(printAct);

   editToolBar = addToolBar(tr("Edit"));
   editToolBar->addAction(cutAct);
   editToolBar->addAction(copyAct);
   editToolBar->addAction(pasteAct);
   editToolBar->addSeparator();
   selectFromSpinBox = new QSpinBox();
   selectFromSpinBox->setToolTip("Select sequence from");
   selectFromSpinBox->setMinimum(1);
   selectFromSpinBox->setMaximum(999999999);
   selectFromSpinBox->setMaximumWidth(50);
   selectToSpinBox = new QSpinBox();
   selectToSpinBox->setToolTip("Select sequence to");
   selectToSpinBox->setMinimum(1);
   selectToSpinBox->setMaximum(999999999);
   selectToSpinBox->setMaximumWidth(50);
   connect(selectFromSpinBox, SIGNAL(valueChanged(int)), this, SLOT(selectionChanged()));
   connect(selectToSpinBox, SIGNAL(valueChanged(int)), this, SLOT(selectionChanged()));
   editToolBar->addAction("From:");
   editToolBar->addWidget(selectFromSpinBox);
   editToolBar->addAction("To:");
   editToolBar->addWidget(selectToSpinBox);
   editToolBar->addSeparator();
   editToolBar->addAction(undoAct);
   editToolBar->addAction(redoAct);

   viewToolBar = addToolBar(tr("View"));
   viewToolBar->addAction(highlightSeqAct);
   selectSeqFormatComboBox = new QComboBox();
   QStringList seqFormatPatterns;
   seqFormatPatterns << "none" << "FASTA Protein" << "FASTA nucleic acid" << "Alignment protein" << "Alignment nucleic acid" << "EMBL" << "Genebank protein" << "Genebank (full)" << "Swissprot";
   selectSeqFormatComboBox->addItems(seqFormatPatterns);
   connect(selectSeqFormatComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(seqFormatChanges(int)));
   viewToolBar->addWidget(selectSeqFormatComboBox);

   graphToolBar = addToolBar(tr("Graph"));
   graphToolBar->addAction(rotateLeftAct);
   graphToolBar->addAction(rotateRightAct);
   graphToolBar->addSeparator();
   graphToolBar->addAction(zoomOutAct);
   graphToolBar->addAction(zoomOrigAct);
   graphToolBar->addAction(zoomInAct);

   runToolBar = addToolBar(tr("Run"));
   runToolBar->addAction(runAct);
   runToolBar->addAction(addParameterAct);
   runToolBar->addAction(copyCommandlineAct);

   helpToolBar = addToolBar(tr("Help"));
   helpToolBar->addAction(helpAct);
   helpToolBar->addAction(whatsThisAct);
}

/*void kprogramWindow::viewTab()
{
   QAction *action = qobject_cast<QAction *>(sender());
   if (action) {
      kpwMainTab->setTabEnabled(action->data().toInt(), !action->isChecked());
      action->setChecked(!action->isChecked());
   }
}*/

void kprogramWindow::findControls()
{
   QList<QObject *> allObjects = findChildren<QObject *>();
   foreach (QObject *object, allObjects) {
    if (object->objectName() != "" && object->objectName() != "qt_spinbox_lineedit" && object->objectName() != "qt_menubar_ext_button" && object->objectName() != "qt_toolbar_ext_button") {

      QLineEdit *lineEdit = qobject_cast<QLineEdit *>(object);
      if (lineEdit) {
         connect(lineEdit, SIGNAL(textEdited(QString)), this, SLOT(wasModified()));
         if (lineEdit->objectName() == "InputFileLineEdit") kpwInputFileLineEdit = lineEdit;
         else if (lineEdit->objectName() == "InputFileLineEdit2") kpwInputFileLineEdit2 = lineEdit;
         else if (lineEdit->objectName() == "OutputFileLineEdit") kpwOutputFileLineEdit = lineEdit;
      }

      QComboBox *comboBox = qobject_cast<QComboBox *>(object);
      if (comboBox) {
         connect(comboBox, SIGNAL(activated(int)), this, SLOT(wasModified()));
         if (comboBox->objectName() == "ProgramComboBox") kpwProgramComboBox = comboBox;
         else if (comboBox->objectName() == "GeneticCodeComboBox") loadFormats("genetic_codes", comboBox);
         else if (comboBox->objectName() == "FormatComboBox") loadFormats("graph_types", comboBox);
         else if (comboBox->objectName() == "SubstitutionMatrixComboBox") loadFormats("substitution_matrices", comboBox);
         else if (comboBox->objectName() == "CodonUsageFileComboBox") loadFormats("codon_usage_files", comboBox);
      }

      QSpinBox *spinBox = qobject_cast<QSpinBox *>(object);
      if (spinBox)
          connect(spinBox, SIGNAL(valueChanged(int)), this, SLOT(wasModified()));

      QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(object);
      if (doubleSpinBox)
          connect(doubleSpinBox, SIGNAL(valueChanged(double)), this, SLOT(wasModified()));

      QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(object);
      if (abstractButton && !abstractButton->objectName().isEmpty() && !abstractButton->text().isEmpty()) {
         connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(wasModified()));
         if (abstractButton->objectName() =="SequenceAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openSAQForm()));
         else if (abstractButton->objectName() =="AssemblyAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openAssAQForm()));
         else if (abstractButton->objectName() =="OutseqAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openOsAQForm()));
         else if (abstractButton->objectName() =="OutfeatAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openOfAQForm()));
         else if (abstractButton->objectName() =="GraphAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openGAQForm()));
         else if (abstractButton->objectName() =="AlignmentAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openAAQForm()));
         else if (abstractButton->objectName() =="OutfileAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openOAQForm()));
      }

      QGroupBox *groupBox = qobject_cast<QGroupBox *>(object);
      if (groupBox && groupBox->isCheckable())
          connect(groupBox, SIGNAL(clicked(bool)), this, SLOT(wasModified()));

      QTextEdit *textEdit = qobject_cast<QTextEdit *>(object);
      if (textEdit) {
         if (textEdit->objectName() == "InputTextEdit") {
             kpwInputTextEdit = textEdit;
             kpwInputTextEdit->setAcceptRichText(false);
             kpwInputTextEdit->installEventFilter(this);
         } else if (textEdit->objectName() == "InputTextEdit2") {
             kpwInputTextEdit2 = textEdit;
             kpwInputTextEdit2->setAcceptRichText(false);
             kpwInputTextEdit2->installEventFilter(this);
         } else if (textEdit->objectName() == "OutputTextEdit") {
             kpwOutputTextEdit = textEdit;
             kpwOutputTextEdit->setAcceptRichText(false);
             kpwOutputTextEdit->installEventFilter(this);
         }
      }

      QGraphicsView *graphicsView = qobject_cast<QGraphicsView *>(object);
      if (graphicsView && graphicsView->objectName() == "OutputGraphicsView")
         kpwOutputGraphicsView = graphicsView;

      QTabWidget *tabWidget = qobject_cast<QTabWidget *>(object);
      if (tabWidget && tabWidget->objectName() == "MainTab") {
         kpwMainTab = tabWidget;
         connect(kpwMainTab, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));
      }
    }
   }
}

bool kprogramWindow::eventFilter(QObject *object, QEvent *event)
{
    if (event->type() == QEvent::FocusOut) {
        if (object == kpwInputTextEdit) {
            disconnect(kpwInputTextEdit, SIGNAL(copyAvailable(bool)), nullptr, nullptr);
            disconnect(kpwInputTextEdit, SIGNAL(undoAvailable(bool)), nullptr, nullptr);
            disconnect(kpwInputTextEdit, SIGNAL(redoAvailable(bool)), nullptr, nullptr);
        } else if (object == kpwInputTextEdit2) {
            disconnect(kpwInputTextEdit2, SIGNAL(copyAvailable(bool)), nullptr, nullptr);
            disconnect(kpwInputTextEdit2, SIGNAL(undoAvailable(bool)), nullptr, nullptr);
            disconnect(kpwInputTextEdit2, SIGNAL(redoAvailable(bool)), nullptr, nullptr);
        } else if (object == kpwOutputTextEdit) {
            disconnect(kpwOutputTextEdit, SIGNAL(copyAvailable(bool)), nullptr, nullptr);
            disconnect(kpwOutputTextEdit, SIGNAL(undoAvailable(bool)), nullptr, nullptr);
            disconnect(kpwOutputTextEdit, SIGNAL(redoAvailable(bool)), nullptr, nullptr);
        }
        disconnect(cutAct, SIGNAL(triggered()), nullptr, nullptr);
        disconnect(copyAct, SIGNAL(triggered()), nullptr, nullptr);
        disconnect(pasteAct, SIGNAL(triggered()), nullptr, nullptr);
        disconnect(undoAct, SIGNAL(triggered()), nullptr, nullptr);
        disconnect(redoAct, SIGNAL(triggered()), nullptr, nullptr);
    } else if (event->type() == QEvent::FocusIn) {
        if (object == kpwInputTextEdit) {
            currentTextEdit = kpwInputTextEdit;
            selectSeqFormatComboBox->setCurrentIndex(seqFormatInput);
        } else if (object == kpwInputTextEdit2) {
            currentTextEdit = kpwInputTextEdit2;
            selectSeqFormatComboBox->setCurrentIndex(seqFormatInput);
        } else if (object == kpwOutputTextEdit) {
            currentTextEdit = kpwOutputTextEdit;
            selectSeqFormatComboBox->setCurrentIndex(seqFormatOutput);
        }
        if (currentTextEdit != nullptr) {
           cutAct->setEnabled(!currentTextEdit->textCursor().selectedText().isEmpty());
           connect(currentTextEdit, SIGNAL(copyAvailable(bool)), cutAct, SLOT(setEnabled(bool)));
           copyAct->setEnabled(!currentTextEdit->textCursor().selectedText().isEmpty());
           connect(currentTextEdit, SIGNAL(copyAvailable(bool)), copyAct, SLOT(setEnabled(bool)));
           connect(cutAct, SIGNAL(triggered()), currentTextEdit, SLOT(cut()));
           connect(copyAct, SIGNAL(triggered()), currentTextEdit, SLOT(copy()));
           connect(pasteAct, SIGNAL(triggered()), currentTextEdit, SLOT(paste()));
           undoAct->setEnabled(currentTextEdit->document()->isUndoAvailable());
           connect(currentTextEdit, SIGNAL(undoAvailable(bool)), undoAct, SLOT(setEnabled(bool)));
           redoAct->setEnabled(currentTextEdit->document()->isRedoAvailable());
           connect(currentTextEdit, SIGNAL(redoAvailable(bool)), redoAct, SLOT(setEnabled(bool)));
           connect(undoAct, SIGNAL(triggered()), currentTextEdit, SLOT(undo()));
           connect(redoAct, SIGNAL(triggered()), currentTextEdit, SLOT(redo()));
           printAct->setEnabled(true);
           findAct->setEnabled(true);
           findNextAct->setEnabled(true);
           findReplaceAct->setEnabled(true);
           m_findDialog->setTextEdit(currentTextEdit);
           m_findReplaceDialog->setTextEdit(currentTextEdit);
           highlighter->setDocument(currentTextEdit->document());
           selectSeqFormatComboBox->setEnabled(true);
        } else {
           cutAct->setEnabled(false);
           copyAct->setEnabled(false);
           undoAct->setEnabled(false);
           redoAct->setEnabled(false);
           findAct->setEnabled(false);
           findNextAct->setEnabled(false);
           findReplaceAct->setEnabled(false);
           printAct->setEnabled(false);
           selectSeqFormatComboBox->setEnabled(false);
        }
    }
    return false;
}

void kprogramWindow::tabChanged(const int &index)
{
   if (kpwOutputGraphicsView != nullptr && (kpwMainTab->tabText(index) == "O&utput" || kpwMainTab->tabText(index) == "Output")) {
      graphMenu->setEnabled(true);
      graphToolBar->setEnabled(true);
      printAct->setEnabled(true);
   } else {
      graphMenu->setEnabled(false);
      graphToolBar->setEnabled(false);
   }
}

void kprogramWindow::setTextEditFonts()
{
    setTextEditFont(kpwInputTextEdit);
    setTextEditFont(kpwInputTextEdit2);
    setTextEditFont(kpwOutputTextEdit);
}

void kprogramWindow::setTextEditFont(QTextEdit *textEdit)
{
    if (textEdit == nullptr) return;

    QTextCursor currentCursor;
    currentCursor = textEdit->textCursor();
    textEdit->selectAll();
    textEdit->setCurrentFont(ksettings->textEditFont());
    textEdit->document()->setDefaultFont(ksettings->textEditFont());
    if (ksettings->wrapLines()) textEdit->setLineWrapMode(QTextEdit::WidgetWidth);
    else textEdit->setLineWrapMode(QTextEdit::NoWrap);
    textEdit->setTextCursor(currentCursor);
}

void kprogramWindow::loadFormats(const QString &fileName, QWidget *widget)
{
   QFile file(ksettings->kembossData() + fileName);
   if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
      QTextStream in(&file);
      while (!in.atEnd()) {
         QString line = in.readLine();
         qobject_cast<QComboBox *>(widget)->addItem(line);
      }
   }
}

void kprogramWindow::newSettings()
{
   if (maybeSave()) {
      deleteTemporaryFiles();
      if (kpwInputTextEdit != nullptr) {
         kpwInputTextEdit->clear();
         kpwInputTextEdit->document()->setModified(false);
      }
      if (kpwInputTextEdit2 != nullptr) {
         kpwInputTextEdit2->clear();
         kpwInputTextEdit2->document()->setModified(false);
      }
      if (kpwOutputTextEdit != nullptr) {
         kpwOutputTextEdit->clear();
      }
      // Set DefaultValues
      if (QFile::exists(defaultSettingsFile()))
         loadSettings(defaultSettingsFile());
      setCurrentFile("");
   }
}

bool kprogramWindow::loadSettings(const QString &fileName)
{
   statusBar()->showMessage(tr("Opening settings"));

   QFile file(fileName);
   if (!file.open(QFile::ReadOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                    tr("Cannot read file %1:\n%2.")
                    .arg(fileName)
                    .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }

   // Load Settings
   QSettings settings(fileName, QSettings::IniFormat);

   QList<QWidget *> allWidgets = findChildren<QWidget *>();
   foreach (QWidget *widget, allWidgets) {
      if (widget->objectName() != "" && widget->objectName() != "qt_spinbox_lineedit" && widget->objectName() != "qt_menubar_ext_button" && widget->objectName() != "qt_toolbar_ext_button" && widget->objectName() !="ProgramComboBox") {
         QLineEdit *lineEdit = qobject_cast<QLineEdit *>(widget);
         if (lineEdit)
            lineEdit->setText(settings.value(lineEdit->objectName()).toString());
         QComboBox *comboBox = qobject_cast<QComboBox *>(widget);
         if (comboBox)
            comboBox->setCurrentIndex(comboBox->findText(settings.value(comboBox->objectName()).toString()));
         QSpinBox *spinBox = qobject_cast<QSpinBox *>(widget);
         if (spinBox)
            spinBox->setValue(settings.value(spinBox->objectName()).toInt());
         QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(widget);
         if (doubleSpinBox)
            doubleSpinBox->setValue(settings.value(doubleSpinBox->objectName()).toDouble());
         QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(widget);
         if (abstractButton && abstractButton->isCheckable())
            abstractButton->setChecked(settings.value(abstractButton->objectName()).toBool());
         QGroupBox *groupBox = qobject_cast<QGroupBox *>(widget);
         if (groupBox && groupBox->isCheckable())
            groupBox->setChecked(settings.value(groupBox->objectName()).toBool());
      }
   }

   associatedQualifiers.clear();
   associatedQualifiers+=settings.value("associated_qualifiers").toStringList();
   additionalConfigurationParameters.clear();
   additionalConfigurationParameters+=settings.value("additional_configuration_parameters").toStringList();
   statusBar()->clearMessage();

   // Open input files if specified
   if (kpwInputFileLineEdit != nullptr && kpwInputFileLineEdit->text() != "")
      loadInputFile(kpwInputTextEdit, kpwInputFileLineEdit->text());
   if (kpwInputFileLineEdit2 != nullptr && kpwInputFileLineEdit2->text() != "")
      loadInputFile(kpwInputTextEdit2, kpwInputFileLineEdit2->text());

   return true;
}

bool kprogramWindow::saveSettings(const QString &fileName)
{
   statusBar()->showMessage(tr("Saving settings"));
   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                   tr("Cannot write file %1:\n%2.")
                   .arg(fileName)
                   .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }

   // Save Settings
   QSettings settings(fileName, QSettings::IniFormat);
   settings.setValue("kemboss_version",QString(KEMBOSS_MAJOR_VERSION)+"."+QString(KEMBOSS_MINOR_VERSION)+"."+QString(KEMBOSS_MINIMUM_VERSION));

   QList<QWidget *> allWidgets = findChildren<QWidget *>();
   foreach (QWidget *widget, allWidgets) {
      if (widget->objectName() != "" && widget->objectName() != "qt_spinbox_lineedit" && widget->objectName() != "qt_menubar_ext_button" && widget->objectName() != "qt_toolbar_ext_button") {
         QLineEdit *lineEdit = qobject_cast<QLineEdit *>(widget);
         if (lineEdit)
             settings.setValue(lineEdit->objectName(),lineEdit->text());
         QComboBox *comboBox = qobject_cast<QComboBox *>(widget);
         if (comboBox)
             settings.setValue(comboBox->objectName(),comboBox->currentText());
         QSpinBox *spinBox = qobject_cast<QSpinBox *>(widget);
         if (spinBox)
             settings.setValue(spinBox->objectName(),spinBox->value());
         QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(widget);
         if (doubleSpinBox)
             settings.setValue(doubleSpinBox->objectName(),doubleSpinBox->value());
         QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(widget);
         if (abstractButton && abstractButton->isCheckable())
             settings.setValue(abstractButton->objectName(),abstractButton->isChecked());
         QGroupBox *groupBox = qobject_cast<QGroupBox *>(widget);
         if (groupBox && groupBox->isCheckable())
             settings.setValue(groupBox->objectName(),groupBox->isChecked());
      }
   }
   if (!associatedQualifiers.isEmpty())
       settings.setValue("associated_qualifiers",associatedQualifiers);
   if (!additionalConfigurationParameters.isEmpty())
       settings.setValue("additional_configuration_parameters",additionalConfigurationParameters);

   setCurrentFile(fileName);
   statusBar()->clearMessage();
   return true;
}

void kprogramWindow::setCurrentFile(const QString &fileName)
{
   if (fileName == defaultSettingsFile()) curFile.clear();
   else curFile = fileName;
   setWindowModified(false); saveAct->setEnabled(false);
   QString shownName;
   if (curFile.isEmpty())
      shownName = "untitled."+objectName();
   else {
      shownName = strippedName(curFile);
      QSettings *settings;
      if (ksettings->portableApp()) {
          settings = new QSettings(ksettings->kembossPortableConfigDir()+objectName(), QSettings::IniFormat);
      } else {
          settings = new QSettings(QSettings::UserScope, "kemboss.org", objectName());
      }
      QStringList files = settings->value("recentFileList").toStringList();
      files.removeAll(fileName);
      files.prepend(fileName);
      while (files.size() > MaxRecentFiles)
         files.removeLast();
      settings->setValue("recentFileList", files);

      foreach (kembossWindow *window, qobject_cast<kemboss *>(kembossParent)->programWindows) {
         if (window->objectName() == objectName())
            window->updateRecentFileActions();

      ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
      }
   }
   setWindowTitle(tr("%1[*] - %2").arg(shownName).arg(objectName()));
}

bool kprogramWindow::on_OpenInputFilePushButton_clicked()
{
   return openInputFilePushButtonClicked(kpwInputTextEdit, kpwInputFileLineEdit);
}

bool kprogramWindow::on_OpenInputFilePushButton2_clicked()
{
   return openInputFilePushButtonClicked(kpwInputTextEdit2, kpwInputFileLineEdit2);
}

bool kprogramWindow::openInputFilePushButtonClicked(QTextEdit *inputTextEdit, QLineEdit *inputFileLineEdit)
{
    if (!maybeSaveInputFiles())
       return false;
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open input file"),ksettings->currentDir(),"All files (*)");
    if (!fileName.isEmpty()) {
       if (inputFileLineEdit != nullptr)
          inputFileLineEdit->setText(fileName);
       ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
       loadInputFile(inputTextEdit, fileName);
    } else return false;
    return true;
}

bool kprogramWindow::loadInputFile(QTextEdit *inputTextEdit, const QString &fileName)
{
   if (inputTextEdit == nullptr)
       return false;

   statusBar()->showMessage(tr("Opening input file"));

   inputTextEdit->clear();

   selectSeqFormat(fileName);

   loadFileThread *loadFile = new loadFileThread(this, fileName, inputTextEdit);
   connect(loadFile, SIGNAL(fileRead(QTextEdit*, QString)), this, SLOT(fileRead(QTextEdit*, QString)));
   connect(loadFile, SIGNAL(finishedLoading(QTextEdit*)), this, SLOT(finishedFileLoading(QTextEdit*)));
   connect(loadFile, SIGNAL(finishedLoading(QTextEdit*)), loadFile, SLOT(deleteLater()));
   loadFile->start(QThread::LowPriority);

   return true;
}

void kprogramWindow::selectSeqFormat(const QString &fileName)
{
    QString suffix = QFileInfo(fileName).suffix();
    if (suffix == "fasta")
       selectSeqFormatComboBox->setCurrentIndex(1);
    else if ((suffix == "alignment") || (suffix == "align") || (suffix == "aln"))
       selectSeqFormatComboBox->setCurrentIndex(3);
    else if (suffix == "embl")
       selectSeqFormatComboBox->setCurrentIndex(5);
    else if ((suffix == "gb") || (suffix == "genebank"))
       selectSeqFormatComboBox->setCurrentIndex(6);
    else if ((suffix == "sp") || (suffix == "swissprot") || (suffix == "swiss"))
       selectSeqFormatComboBox->setCurrentIndex(8);
    else
       selectSeqFormatComboBox->setCurrentIndex(0);
}

void kprogramWindow::fileRead(QTextEdit *textEdit, QString text)
{
    qDebug() << "Loaded file";
    textEdit->append(text);
}

void kprogramWindow::finishedFileLoading(QTextEdit *textEdit)
{
    qDebug() << "Finished file loading";
    textEdit->document()->setModified(false);
    statusBar()->clearMessage();
}

bool kprogramWindow::on_SaveInputFileAsPushButton_clicked()
{
    return saveInputFileAsPushButtonClicked(kpwInputTextEdit, kpwInputFileLineEdit);
}

bool kprogramWindow::on_SaveInputFileAsPushButton2_clicked()
{
    return saveInputFileAsPushButtonClicked(kpwInputTextEdit2, kpwInputFileLineEdit2);
}

bool kprogramWindow::saveInputFileAsPushButtonClicked(QTextEdit *inputTextEdit, QLineEdit *inputFileLineEdit)
{
    if (inputTextEdit != nullptr) {
       QString fileName = QFileDialog::getSaveFileName(this, tr("Save input file"),ksettings->currentDir(),"All files (*)");
       if (!fileName.isEmpty() && saveInputFile(inputTextEdit, fileName)) {
          if (inputFileLineEdit != nullptr) inputFileLineEdit->setText(fileName);
          qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
       } else return false;
    } else return false;
    return true;
}

bool kprogramWindow::saveInputFile(QTextEdit *inputTextEdit, const QString &fileName)
{
   statusBar()->showMessage(tr("Saving input file"));
   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot save input file %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }
   QTextStream out(&file);
   out << inputTextEdit->toPlainText();
   inputTextEdit->document()->setModified(false);
   statusBar()->clearMessage();
   return true;
}

bool kprogramWindow::on_SaveOutputFileAsPushButton_clicked()
{
   QString fileName = QFileDialog::getSaveFileName(this, tr("Save output"),ksettings->currentDir(),"All files (*)");
   if (fileName.isEmpty())
       return false;

   if (kpwOutputTextEdit != nullptr && kpwOutputTextEdit->document()->isModified()) {
   // The contents of the output text edit as been modified by the user
      saveOutputFile(fileName);
   } else if (!tempOutputFile.isEmpty() && QFile::exists(tempOutputFile)) {
   // The output of the EMBOSS program has been saved to a temporary file but has not been changed by the user
      // If the selected file already exists, delete it
      if (QFile::exists(fileName))
          QFile::remove(fileName);
      QFile::rename(tempOutputFile,fileName);
      tempOutputFile.clear();
   }
   if (kpwOutputFileLineEdit != nullptr)
      kpwOutputFileLineEdit->setText(fileName);
   ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
   return true;
}

bool kprogramWindow::saveOutputFile(const QString &fileName)
{
   statusBar()->showMessage(tr("Saving output file"));
   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot save output file %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }
   QTextStream out(&file);
   out << kpwOutputTextEdit->toPlainText();
   kpwOutputTextEdit->document()->setModified(false);
   file.close();

   if (QFile::exists(tempOutputFile)) {
   // If a temporary EMBOSS output file exists, it has to be deleted
      tempOutputFile.clear();
   }
   statusBar()->clearMessage();
   return true;
}

void kprogramWindow::loadPrinterSettings()
{
    // Load printer settings
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"Configuration", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "Configuration");
    }
    printer.setCollateCopies(settings->value("print_collate_copies").toBool());
    printer.setColorMode(QPrinter::ColorMode(settings->value("print_color_mode").toInt()));
    printer.setDuplex(QPrinter::DuplexMode(settings->value("print_duplex_mode").toInt()));
    printer.setCreator(objectName());
    printer.setFontEmbeddingEnabled(settings->value("print_font_embedding_enabled").toBool());
    printer.setOrientation(QPrinter::Orientation(settings->value("print_orientation").toInt()));
    printer.setOutputFormat(QPrinter::OutputFormat(settings->value("print_output_format").toInt()));
    printer.setPageMargins(settings->value("print_margin_left").toDouble(),
                           settings->value("print_margin_top").toDouble(),
                           settings->value("print_margin_right").toDouble(),
                           settings->value("print_margin_bottom").toDouble(),
                           QPrinter::Unit(settings->value("print_unit").toInt()));
    printer.setPageOrder(QPrinter::PageOrder(settings->value("print_page_order").toInt()));
    printer.setPaperSize(QPrinter::PaperSize(settings->value("print_paper_size").toInt()));
    printer.setPaperSource(QPrinter::PaperSource(settings->value("print_paper_source").toInt()));
    if (!settings->value("print_printer_name").toString().isEmpty())
        printer.setPrinterName(settings->value("print_printer_name").toString());
    printer.setResolution(settings->value("print_resolution").toInt());
}

void kprogramWindow::savePrinterSettings()
{
    // Save the current settings of the printer object
    QSettings *settings;
    if (ksettings->portableApp()) {
        settings = new QSettings(ksettings->kembossPortableConfigDir()+"Configuration", QSettings::IniFormat);
    } else {
        settings = new QSettings(QSettings::UserScope, "kemboss.org", "Configuration");
    }
    settings->setValue("print_collate_copies", printer.collateCopies());
    settings->setValue("print_color_mode", printer.colorMode());
    settings->setValue("print_duplex_mode", printer.duplex());
    settings->setValue("print_font_embedding_enabled", printer.fontEmbeddingEnabled());
    qreal pleft, ptop, pright, pbottom;
    QPrinter::Unit unit = QPrinter::Millimeter;
    printer.getPageMargins(&pleft, &ptop, &pright, &pbottom, unit);
    settings->setValue("print_margin_left", pleft);
    settings->setValue("print_margin_top", ptop);
    settings->setValue("print_margin_right", pright);
    settings->setValue("print_margin_bottom", pbottom);
    settings->setValue("print_unit", unit);
    settings->setValue("print_orientation", printer.orientation());
    settings->setValue("print_output_format", printer.outputFormat());
    settings->setValue("print_page_order", printer.pageOrder());
    settings->setValue("print_paper_size", printer.paperSize());
    settings->setValue("print_paper_source", printer.paperSource());
    settings->setValue("print_printer_name", printer.printerName());
    settings->setValue("print_resolution", printer.resolution());
}

void kprogramWindow::printDocument()
{
#ifndef QT_NO_PRINTER
   QPrintDialog *dialog = new QPrintDialog(&printer, this);
   dialog->setWindowTitle(tr("Print Document"));
   if (currentTextEdit != nullptr && currentTextEdit->textCursor().hasSelection())
      dialog->addEnabledOption(QAbstractPrintDialog::PrintSelection);
   if (dialog->exec() != QDialog::Accepted)
      return;
   savePrinterSettings();

   if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "O&utput" && kpwOutputGraphicsView != nullptr) {
       QPainter painter(&printer);
       kpwOutputGraphicsView->render(&painter);
   } else if (currentTextEdit != nullptr) {
       currentTextEdit->print(&printer);
   }
#else
   QMessageBox::information(this, tr("Printing canceled"), tr("Printing is not supported on this Qt build"), QMessageBox::Cancel);
#endif
}

void kprogramWindow::documentProperties()
{
   QPageSetupDialog *dialog = new QPageSetupDialog(&printer, this);
   if (dialog->exec() == QDialog::Accepted)
      savePrinterSettings();
}

void kprogramWindow::seqFormatNone()
{
   selectSeqFormatComboBox->setCurrentIndex(0);
}

void kprogramWindow::seqFormatFastaAA()
{
   selectSeqFormatComboBox->setCurrentIndex(1);
}

void kprogramWindow::seqFormatFastaNuc()
{
   selectSeqFormatComboBox->setCurrentIndex(2);
}

void kprogramWindow::seqFormatAlignmentAA()
{
   selectSeqFormatComboBox->setCurrentIndex(3);
}

void kprogramWindow::seqFormatAlignmentNuc()
{
   selectSeqFormatComboBox->setCurrentIndex(4);
}

void kprogramWindow::seqFormatEmbl()
{
   selectSeqFormatComboBox->setCurrentIndex(5);
}

void kprogramWindow::seqFormatGbAA()
{
   selectSeqFormatComboBox->setCurrentIndex(6);
}

void kprogramWindow::seqFormatGbNuc()
{
   selectSeqFormatComboBox->setCurrentIndex(7);
}

void kprogramWindow::seqFormatSwiss()
{
   selectSeqFormatComboBox->setCurrentIndex(8);
}

void kprogramWindow::seqFormatChanges(int index)
{
   if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "&Input") {
      seqFormatInput = index;
   } else if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "O&utput") {
      seqFormatOutput = index;
   }
   if (highlightSeqAct->isChecked()) {
       highlighter->setSyntaxInt(index);
       highlighter->rehighlight();
   }
}

void kprogramWindow::highlightSeq(bool highlight)
{
   if (highlight) {
       highlighter->setSyntaxInt(selectSeqFormatComboBox->currentIndex());
       highlighter->rehighlight();
   } else {
       highlighter->setSyntaxInt(0);
       highlighter->rehighlight();
   }
}

void kprogramWindow::selectionChanged()
{
    // The select to value always has to be higher than the select from value
    if (selectToSpinBox->value() <= selectFromSpinBox->value())
        selectToSpinBox->setValue(selectFromSpinBox->value());

    // Find start and end position of sequence in document
    QRegExp seqStartRegExp;
    QRegExp seqEndRegExp;
    bool seqStartsAtEndOfRegExp = false; // True if the sequence starts at the end of the RegExp, false if a match of the RegExp is the start of the sequence

 //   bool seqIsAA; // Is set true if the sequence is AA, otherwise is set false

    switch (selectSeqFormatComboBox->currentIndex()) {
    case 0: // None
        // The start of the sequence is the first word character
        seqStartRegExp = QRegExp("^\\w+");
        seqStartsAtEndOfRegExp = false;
     //   seqIsAA = true;
        break;
    case 2: // FastaNuc
        // The start of the sequence is the first line that does not start with a > (Sequence name), a # (Comment), or a / (Feature)
        seqStartRegExp = QRegExp("^[^\\>#/].*$");
        seqStartsAtEndOfRegExp = false;
     //   seqIsAA = false;
        break;
    case 1: // FastaAA
        // The start of the sequence is the first line that does not start with a > (Sequence name), a # (Comment), or a / (Feature)
        seqStartRegExp = QRegExp("^[^\\>#/].*$");
        seqStartsAtEndOfRegExp = false;
      //  seqIsAA = true;
        break;
    case 3: // AlignmentAA
        // The start of the sequence is the first line that does not start with CLUSTAL (Alignment title), with a > (Sequence name), or an alignment score
        seqStartRegExp = QRegExp("^[^(CLUSTAL|\\>|Score|Identities| Plus| Minus)].*$");
        seqStartsAtEndOfRegExp = false;
    //    seqIsAA = true;
        break;
    case 4: // AlignmentNuc
        // The start of the sequence is the first line that does not start with CLUSTAL (Alignment title), with a > (Sequence name), or an alignment score
        seqStartRegExp = QRegExp("^[^(CLUSTAL|\\>|Score|Identities| Plus| Minus)].*$");
        seqStartsAtEndOfRegExp = false;
     //   seqIsAA = false;
        break;
    case 5: // EMBL
        // The start of the sequence is the first line that matches one ore more whitespaces, one or more word characters or whitespaces, then one or more digits, end of line
        seqStartRegExp = QRegExp("^\\s+[\\w\\s]+\\d+$");
        seqEndRegExp = QRegExp("//");
        seqStartsAtEndOfRegExp = false;
     //   seqIsAA = true;
        break;
    case 6: // GenbankAA
        seqStartRegExp = QRegExp("ORIGIN\\s");
        seqStartsAtEndOfRegExp = true;
      //  seqIsAA = true;
        break;
    case 7: // GenbankNuc
        seqStartRegExp = QRegExp("ORIGIN\\s");
        seqStartsAtEndOfRegExp = true;
      //  seqIsAA = false;
        break;
    case 8: //Swiss
        // The start of the sequence is the first line that starts with at least 3 word characters
        seqStartRegExp = QRegExp("^\\w{3,}");
        seqEndRegExp = QRegExp("//");
        seqStartsAtEndOfRegExp = false;
      //  seqIsAA = true;
        break;
    }

    QTextCursor found;
    int seqStart;
    int seqEnd;

    found = currentTextEdit->document()->find(seqStartRegExp, 0); // Find start position of sequence in document
    if (!found.isNull()) {
        if (seqStartsAtEndOfRegExp)
            seqStart = found.selectionEnd();
        else
            seqStart = found.selectionStart();

        if (seqEndRegExp.isEmpty()) {
            // If the end of the sequence is not defined, the end of the document is regarded as the end of the sequence
            seqEnd = currentTextEdit->document()->characterCount();
        } else {
            found = currentTextEdit->document()->find(seqEndRegExp, seqStart); // Find end position of sequence in document
            if (found.isNull()) {
                seqEnd = currentTextEdit->document()->characterCount();
            } else {
                seqEnd = found.selectionStart();
            }
        }
    } else {
        // The start position of the sequence has not been found, no selection is performed
        return;
    }

    // Find start position of selected sequence in document
    int selectedSeqStart = 0;
    int foundSeqLength = 0;
    int searchLength = 0;

    while ((foundSeqLength < selectFromSpinBox->value()) && (seqStart+searchLength < seqEnd)) {
        if (charIsAminoacid(currentTextEdit->document()->characterAt(seqStart+searchLength))) {
            ++foundSeqLength;
        }
        ++searchLength;
    }
    selectedSeqStart = seqStart + searchLength - 1;

    // Find end position of selected sequence in document
    foundSeqLength = 0;
    int selectedSeqEnd;

    while ((foundSeqLength < (selectToSpinBox->value() - selectFromSpinBox->value())) && (seqStart+searchLength < seqEnd)) {
        if (charIsAminoacid(currentTextEdit->document()->characterAt(seqStart+searchLength))) {
            ++foundSeqLength;
        }
        ++searchLength;
    }
    selectedSeqEnd = seqStart + searchLength;

    // Select sequence
    QTextCursor selection;
    selection = currentTextEdit->textCursor();
    selection.setPosition(selectedSeqStart, QTextCursor::MoveAnchor);
    selection.setPosition(selectedSeqEnd, QTextCursor::KeepAnchor);
    currentTextEdit->setTextCursor(selection);
}

bool kprogramWindow::charIsAminoacid(const QChar &seq)
{
    QRegExp rx("[GCAVFPMILSTYHNQWDEKRgcavfpmilstyhnqtwdekr]");
    return rx.exactMatch(seq);
}

bool kprogramWindow::charIsNucleotide(const QChar &seq)
{
    QRegExp rx("[ATGCNatgcn]");
    return rx.exactMatch(seq);
}

bool kprogramWindow::sequenceIsNucleotide(const QString &seq)
{
    QString testString = seq;
    testString.truncate(100);
    testString.remove(QRegExp("\\W")); // Removes any non-word characters from the string
    int atgcnCount = testString.count(QRegExp("[ATGCNatgcn]"));
    if ((atgcnCount / testString.length()) > 0.9)
        return true;
    else
        return false;
}

void kprogramWindow::find()
{
   m_findDialog->show();
}

void kprogramWindow::findReplace()
{
   m_findReplaceDialog->show();
}

void kprogramWindow::zoomIn()
{
   kpwOutputGraphicsView->scale(1.2, 1.2);
   graphScale *= 1.2;
}

void kprogramWindow::zoomOut()
{
   kpwOutputGraphicsView->scale(0.8, 0.8);
   graphScale *= 0.8;
}

void kprogramWindow::zoomOriginal()
{
   kpwOutputGraphicsView->scale(1/graphScale, 1/graphScale);
   graphScale = 1;
}

void kprogramWindow::rotateLeft()
{
   kpwOutputGraphicsView->rotate(-90);
}

void kprogramWindow::rotateRight()
{
   kpwOutputGraphicsView->rotate(90);
}

bool kprogramWindow::runEmboss()
{
   deleteTemporaryFiles();
   if (!checkInputs()) return false;
   if (!checkOutput()) return false;
   beforeRunningProgram();
   if (!createProgramConfiguration()) return false;
   if (!runEmbossApp->runApp()) return false;
   return true;
}

bool kprogramWindow::checkInputs()
{
    runEmbossApp->clearEmbossInputFileList();

    if (!checkInput(kpwInputTextEdit, kpwInputFileLineEdit))
        return false;
    if (!checkInput(kpwInputTextEdit2, kpwInputFileLineEdit2))
        return false;

    return true;
}

bool kprogramWindow::checkInput(QTextEdit *inputTextEdit, QLineEdit *inputFileLineEdit)
{
   if (inputFileLineEdit == nullptr) return true;

   // If no input file is specified
   if (inputFileLineEdit->text().isEmpty()) {
       //  If no text has been entered in the InputFileTextEdit, no input data is available
       if (inputTextEdit == nullptr || inputTextEdit->document()->isEmpty()) {
            QMessageBox::warning(this, objectName(),
                       tr("No input available!\n"
                       "Please, either select an appropriate input file or enter Your input data into the corresponding text box."));
            return false;
        //  If text has been entered in the InputFileTextEdit, create a temporary input file
        } else
            saveTextToTempfile(inputTextEdit, inputFileLineEdit);

   // If the text in the InputFileTextEdit has changed since the input file has been opened, ask what to do
   } else if (inputTextEdit != nullptr && inputTextEdit->document()->isModified()) {
       QMessageBox::StandardButton ret;
       ret = QMessageBox::warning(this, objectName(),
                         tr("The input file has been modified.\n"
                         "Do you want to save your changes before running the emboss application?"),
                         QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
       if (ret == QMessageBox::Save) {
           if (!saveInputFile(inputTextEdit, inputFileLineEdit->text()))
               return false;
       } else if (ret == QMessageBox::Cancel) {
           return false;
       }
   }

   // Add the EMBOSS input file to the embossInputFiles list, to transfer the input file to the server if EMBOSS is executed remotely
   runEmbossApp->addEmbossInputFile(inputFileLineEdit->text());

   return true;
}

bool kprogramWindow::saveTextToTempfile(QTextEdit *textEdit, QLineEdit *lineEdit)
{
    if (!tempDir->isValid())
        return false;

    QString fileIndex;
    fileIndex.setNum(tempInputFiles.length());

    tempInputFiles.append(tempDir->path() + "/" + runEmbossApp->embossApp() + fileIndex + ".input");

    if (!saveInputFile(textEdit, tempInputFiles[tempInputFiles.length()-1]))
        return false;

    lineEdit->setText(tempInputFiles[tempInputFiles.length()-1]);
    textEdit->document()->setModified();
    return true;
}

bool kprogramWindow::checkOutput()
{
    runEmbossApp->clearEmbossOutputFileList();
    if (kpwOutputFileLineEdit != nullptr) {

      // If the output file is not specified create a temporary directory containing a temporary output file
      if (kpwOutputFileLineEdit->text().isEmpty()) {
          if (tempDir->isValid()) {
              tempOutputFile = tempDir->path() + "/" + runEmbossApp->embossApp() + ".output";
          }
          kpwOutputFileLineEdit->setText(tempOutputFile);

      // Delete file if output file already exists and auto-delete option is enabled. Otherwise ask what to do
      } else {
         if (QFile::exists(kpwOutputFileLineEdit->text())) {
            if (ksettings->autoDeleteExistingOutputFile())
               QFile::remove(kpwOutputFileLineEdit->text());
            else {
               QMessageBox::StandardButton ret;
               ret = QMessageBox::warning(this, objectName(),
                  tr("The output file already exists.\n"
                  "Delete current output file?"),
                  QMessageBox::Yes | QMessageBox::Ignore | QMessageBox::Cancel);
               if (ret == QMessageBox::Yes)
                  QFile::remove(kpwOutputFileLineEdit->text());
               else if (ret == QMessageBox::Cancel)
                  return false;
            }
         }
      }

      runEmbossApp->addEmbossOutputFile(kpwOutputFileLineEdit->text());
    }
    if (kpwOutputTextEdit != nullptr) kpwOutputTextEdit->clear();
    return true;
}

bool kprogramWindow::createProgramConfiguration()
{
    runEmbossApp->programConfiguration.clear();
    runEmbossApp->addConfigParam("-auto");

    QList<QWidget *> allWidgets = findChildren<QWidget *>();
    foreach (QWidget *widget, allWidgets) {
        if (!widget->toolTip().isEmpty() && widget->isEnabled()) {
            QLineEdit *lineEdit = qobject_cast<QLineEdit *>(widget);
            if (lineEdit && !lineEdit->text().isEmpty())
                runEmbossApp->addConfigParam("-"+lineEdit->toolTip()+"="+lineEdit->text());

            QComboBox *comboBox = qobject_cast<QComboBox *>(widget);
            if (comboBox) {
                if (!(comboBox->currentText()=="default"))
                    runEmbossApp->addConfigParam("-"+comboBox->toolTip()+"="+comboBox->currentText());
            }

            QSpinBox *spinBox = qobject_cast<QSpinBox *>(widget);
            if (spinBox && !(spinBox == selectFromSpinBox) && !(spinBox == selectToSpinBox)) runEmbossApp->addConfigParam("-"+spinBox->toolTip()+"="+spinBox->text());

            QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(widget);
            if (doubleSpinBox) {
                QString num;
                num.setNum(doubleSpinBox->value());
                runEmbossApp->addConfigParam("-"+doubleSpinBox->toolTip()+"="+num);
            }

            QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(widget);
            if (abstractButton && abstractButton->isCheckable()) {
                if (abstractButton->toolTip().startsWith("[no]")) {
                    if (abstractButton->isChecked())
                        runEmbossApp->addConfigParam("-"+abstractButton->toolTip().remove(0,4));
                    else runEmbossApp->addConfigParam("-"+abstractButton->toolTip().replace(0,4,"no"));
                } else {
                    if (abstractButton->isChecked() && abstractButton->toolTip() != "Highlight sequence")
                        runEmbossApp->addConfigParam("-"+abstractButton->toolTip());
                }
            }

            QGroupBox *groupBox = qobject_cast<QGroupBox *>(widget);
            if (groupBox && groupBox->isCheckable()) {
                if (groupBox->toolTip().startsWith("[no]")) {
                    if (!groupBox->isChecked())
                        runEmbossApp->addConfigParam("-"+groupBox->toolTip().replace(0,4,"no"));
                } else {
                    if (groupBox->isChecked())
                        runEmbossApp->addConfigParam("-"+groupBox->toolTip());
                }
            }
        }
    }
    runEmbossApp->programConfiguration += associatedQualifiers;
    runEmbossApp->programConfiguration += additionalConfigurationParameters;

    runEmbossApp->programConfiguration.replaceInStrings("EMBOSS_TEST", ksettings->embossTest(), Qt::CaseSensitive);

    return true;
}

void kprogramWindow::embossExecutionFinished(const int exitCode)
{
    if (exitCode == 0)
        displayOutputFile();
}

void kprogramWindow::completeOutputFilename(QLineEdit *lineEdit, const QString &extension)
{
   if (lineEdit->text().isEmpty() && !kpwOutputFileLineEdit->text().isEmpty())
      lineEdit->setText(kpwOutputFileLineEdit->text()+"."+extension);
   if (!lineEdit->text().isEmpty())
       runEmbossApp->addEmbossOutputFile(lineEdit->text());
}

bool kprogramWindow::displayOutputFile()
{
   // Try to display the program output on the 'Output' tab
    if (kpwOutputFileLineEdit == nullptr)
        return false;

    statusBar()->showMessage(tr("Opening output file"));

    if (kpwOutputTextEdit != nullptr && QFile::exists(kpwOutputFileLineEdit->text())) {
        QFile file(kpwOutputFileLineEdit->text());
        if (file.open(QFile::ReadOnly | QFile::Text)) {
            QTextStream in(&file);
            kpwOutputTextEdit->setPlainText(in.readAll());
        }
        kpwOutputTextEdit->document()->setModified(false);
    } else if (kpwOutputGraphicsView != nullptr) {
        loadOutputGraphics(kpwOutputFileLineEdit->text());
    }
    if (ksettings->autoShowOutputTab() && kpwMainTab != nullptr) {
        if (kpwMainTab->tabText(1) == "O&utput")
            kpwMainTab->setCurrentIndex(1);
        else if (kpwMainTab->tabText(0) == "O&utput")
            kpwMainTab->setCurrentIndex(0);
    }
    statusBar()->clearMessage();
    return true;
}

void kprogramWindow::loadOutputGraphics(const QString &filename)
{
    // Search for possibly created graphical output
    QString graphicsFile = "";
    if (QFile::exists(filename+".1.png"))
        graphicsFile = filename+".1.png";
    else if (QFile::exists(filename+".gif"))
        graphicsFile = filename+".gif";
    else if (QFile::exists(filename+".pdf"))
        QDesktopServices::openUrl(QUrl::fromLocalFile(filename+".pdf"));
    else if (QFile::exists(filename+".svg"))
        QDesktopServices::openUrl(QUrl::fromLocalFile(filename+".svg"));

    // Try to open the graphical output
    if (!graphicsFile.isEmpty()) {
        QGraphicsScene *scene;
        scene = new QGraphicsScene(this);
        kpwOutputGraphicsView->setScene(scene);
        QPixmap pixmap;
        if (pixmap.load(graphicsFile)) {
            scene->addPixmap(pixmap);
            graphScale = 1;
        } else QMessageBox::warning(this, objectName(), tr("Couldn't load the output graphics ")+graphicsFile);
    }
}

void kprogramWindow::openSAQForm(const QString num)
{
   SAQForm newSAQForm(this, num);
   newSAQForm.exec();
}

void kprogramWindow::openAssAQForm(const QString num)
{
   AssemblyAQForm newAssemblyAQForm(this, num);
   newAssemblyAQForm.exec();
}

void kprogramWindow::openOAQForm(const QString num)
{
   OAQForm newOAQForm(this, num);
   newOAQForm.exec();
}

void kprogramWindow::openOsAQForm(const QString num)
{
   OsAQForm newOsAQForm(this, num);
   newOsAQForm.exec();
}

void kprogramWindow::openOfAQForm(const QString num)
{
   OfAQForm newOfAQForm(this, num);
   newOfAQForm.exec();
}

void kprogramWindow::openFAQForm(const QString num)
{
   FAQForm newFAQForm(this, num);
   newFAQForm.exec();
}

void kprogramWindow::openGAQForm(const QString num)
{
   GAQForm newGAQForm(this, num);
   newGAQForm.exec();
}

void kprogramWindow::openAAQForm(const QString num)
{
   AAQForm newAAQForm(this, num);
   newAAQForm.exec();
}

void kprogramWindow::addParameter()
{
   ACPForm newACPForm(this);
   newACPForm.exec();
}

void kprogramWindow::copyCommandline()
{
   if (createProgramConfiguration())
      QApplication::clipboard()->setText(ksettings->embossBin()+runEmbossApp->embossApp()
      #ifdef Q_OS_WIN32
      +".exe"
      #endif
      +" "+runEmbossApp->programConfiguration.join(" "));
}

void kprogramWindow::embossHelp()
{
   assistant(runEmbossApp->embossApp());
}

QString kprogramWindow::selectOpenFile(const QString &msg, QLineEdit *lineEdit)
{
   QString fileName = QFileDialog::getOpenFileName(this, msg, ksettings->currentDir(),"All files (*)");
   if (!fileName.isEmpty()) {
      lineEdit->setText(fileName);
      ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
   }
   return fileName;
}

void kprogramWindow::selectAndOpenFile(const QString &msg, QLineEdit *lineEdit, QTextEdit *textEdit)
{
    QString fileName = selectOpenFile(msg, lineEdit);
    if (!fileName.isEmpty())
        loadInputFile(textEdit, fileName);
}

QString kprogramWindow::selectSaveFile(const QString &msg, QLineEdit *lineEdit)
{
   QString fileName = QFileDialog::getSaveFileName(this, msg, ksettings->currentDir(),"All files (*)");
   if (!fileName.isEmpty()) {
      lineEdit->setText(fileName);
      ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());
   }
   return fileName;
}

void kprogramWindow::selectAndSaveFile(const QString &msg, QLineEdit *lineEdit, QTextEdit *textEdit)
{
    QString fileName = selectSaveFile(msg, lineEdit);
    if (!fileName.isEmpty()) {
        saveInputFile(textEdit, fileName);
    }
}

QStringList kprogramWindow::loadRebaseEnzymes()
{
   // Load enzyme list from rebase database
   QStringList enzymes;
   QFile file(ksettings->embossData()+"REBASE/embossre.enz");
   if (file.open(QFile::ReadOnly | QFile::Text)) {
      QTextStream in(&file);
      QRegExp sep("\\s+");
      while (!in.atEnd()) {
         QString line = in.readLine();
         if (!(line.startsWith("#") || line.isEmpty()))  // exclude comments and empty lines
            enzymes.append(line.section(sep, 0, 0));
      }
   }
   return enzymes;
}

void kprogramWindow::saveEnzymeList(QListWidget *enzymeList)
{
   QString fileName = QFileDialog::getSaveFileName(this, tr("Save enzyme list"),ksettings->currentDir(),"All files (*)");
   if (fileName.isEmpty()) return;
   ksettings->setcurrentDir(QFileInfo(fileName).absolutePath());

   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot save enzyme list %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      return;
   }
   QTextStream out(&file);
   for (int i = 0; i < enzymeList->count(); ++i) {
      out << enzymeList->item(i)->text()+"\n";
   }
}

QString kprogramWindow::selectedEnzymes(QListWidget *enzymeList)
{
   QString enzymes;
   for (int i = 0; i < enzymeList->count(); ++i)
      enzymes += enzymeList->item(i)->text()+",";
   enzymes.chop(1);
   return "-enzymes="+enzymes;
}

QString kprogramWindow::defaultSettingsFile()
{
   return ksettings->kembossDefaultSettings()+"default."+objectName();
}

bool kprogramWindow::maybeSave()
{
   // Check whether the kemboss settings need to be saved
   if (isWindowModified()) {
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The settings have been modified.\n"
            "Do you want to save your changes?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save) {
         if (!save()) return false;
      } else if (ret == QMessageBox::Cancel) {
         return false;
      }
   }

   if (!maybeSaveInputFiles())
      return false;

   if (!maybeSaveOutputFile())
      return false;

   return true;
}

bool kprogramWindow::maybeSaveInputFiles()
{
    if (kpwInputTextEdit != nullptr)
        if (!maybeSaveInputFile(kpwInputTextEdit, kpwInputFileLineEdit)) return false;
    if (kpwInputTextEdit2 != nullptr)
        if (!maybeSaveInputFile(kpwInputTextEdit2, kpwInputFileLineEdit2)) return false;
    return true;
}

bool kprogramWindow::maybeSaveInputFile(QTextEdit *inputTextEdit, QLineEdit *inputFileLineEdit)
// Check whether the input file has been changed
{
   if (inputTextEdit->document()->isModified()) {
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The input data has been modified.\n"
            "Do you want to save your changes?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save && inputFileLineEdit != nullptr) {
         if (inputFileLineEdit->text().isEmpty()) {
            return saveInputFileAsPushButtonClicked(inputTextEdit, inputFileLineEdit);
         } else {
            return saveInputFile(inputTextEdit, inputFileLineEdit->text());
         }
      } else if (ret == QMessageBox::Cancel) {
         return false;
      }
   }
   return true;
}

bool kprogramWindow::maybeSaveOutputFile()
// Check whether the output file has been changed or whether there is a temporary output file that has to be saved
{
   if (kpwOutputTextEdit != nullptr && kpwOutputTextEdit->document()->isModified()) {
   // The contents of the output text edit as been modified by the user
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The output of the EMBOSS program has been modified.\n"
            "Do you want to save your changes?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save && kpwOutputFileLineEdit != nullptr) {
          if (kpwOutputFileLineEdit->text().isEmpty() || !tempOutputFile.isEmpty()) {
              return on_SaveOutputFileAsPushButton_clicked();
          } else {
              return saveOutputFile(kpwOutputFileLineEdit->text());
          }
      } else if (ret == QMessageBox::Cancel) {
         return false;
      } else if (ret == QMessageBox::Discard) {
         return true;
      }
   }

   // The output of the EMBOSS program has been saved to a temporary file but has not been changed by the user
   if (!tempOutputFile.isEmpty()) {
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The output of the EMBOSS program has not been saved.\n"
            "Do you want to save the program output?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save) {
         return on_SaveOutputFileAsPushButton_clicked();
      } else if (ret == QMessageBox::Cancel) {
         return false;
      }
   }
   return true;
}

void kprogramWindow::deleteTemporaryFiles()
{
    // Delete temporary input files
    for (int i=0; i<tempInputFiles.length(); i++) {
        if (kpwInputFileLineEdit != nullptr && kpwInputFileLineEdit->text() == tempInputFiles[i])
            kpwInputFileLineEdit->clear();
        if (kpwInputFileLineEdit2 != nullptr && kpwInputFileLineEdit2->text() == tempInputFiles[i])
            kpwInputFileLineEdit2->clear();
        QFile::remove(tempInputFiles[i]);
        tempInputFiles[i].clear();
    }

    // Delete temporary output files
    if (!tempOutputFile.isEmpty()) {
        if (kpwOutputFileLineEdit != nullptr && kpwOutputFileLineEdit->text() == tempOutputFile)
            kpwOutputFileLineEdit->clear();
        QFile::remove(tempOutputFile);
        tempOutputFile.clear();
    }
    // The temporary dir will be deleted within the close event
}

void kprogramWindow::closeEvent(QCloseEvent *event)
{
   if (maybeSave()) {
      deleteTemporaryFiles();
      writeSettings();
      QSettings *settings;
      if (ksettings->portableApp()) {
          settings = new QSettings(ksettings->kembossPortableConfigDir()+objectName(), QSettings::IniFormat);
      } else {
          settings = new QSettings(QSettings::UserScope, "kemboss.org", objectName());
      }
      m_findDialog->writeSettings(settings);
      m_findReplaceDialog->writeSettings(settings);
      qobject_cast<kemboss *>(kembossParent)->programWindows.removeAll(this);
      tempDir->remove();
      runEmbossApp->~krunEmbossApp();
      event->accept();
   } else event->ignore();
}
