/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2020 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#ifndef RUNEMBOSSAPPONSERVER_H
#define RUNEMBOSSAPPONSERVER_H

#include <QtNetwork>
#include <QSslSocket>

#include "settings.h"
#include "runEmbossApp.h"

class krunEmbossAppOnServer : public krunEmbossApp
{
   Q_OBJECT

public:
    krunEmbossAppOnServer(QWidget *parent = nullptr, kembossSettings *fksettings = nullptr);
    virtual ~krunEmbossAppOnServer() {}
    virtual bool runApp();

private slots:
    virtual void processFinished(int exitCode);
    virtual void killProcess();

    void sessionOpened();
    void kembossHostFound();
    void connectionEstablished();
    // After the program configuration and the input files have been received, the emboss program can be executed
    void socketEncrypted();
    // This slot is called when an encrypted connection has been established
    void readData();
    // Read all incoming data and return it as a string

    void connectionClosed();

    void proxyAuthenticationRequired (const QNetworkProxy &, QAuthenticator *);
    // kemboss doesn't yet support connection via a network proxy

    void displayError(QAbstractSocket::SocketError error);
    // Display errors in TCP connection

    void checkPeerVerifyError(const QSslError &error);
    // An error has occurred while establishing the identity of the peer

    void checkSslErrors(const QList<QSslError> &errors);
    // Display errors in SSL connection

    void displayCertificateInfo();
    // Display information on the SSL certificate

    void ignoreHostSSLErrors();
    // The hostname is added to a list of hosts whose SSL errors are always ignored

private:
    bool startKembossClient();
    qint64 writeData(QByteArray data);
    // Send data to the server
    void writeUser();
    // Sends the kemboss client version, user name, group, email and password to the server
    void writeEmbossConfiguration();
    // Sends the EMBOSS program name and configuration to the server
    void writeEmbossOutputFileList();
    // Sends the list of EMBOSS output files to the server
    QString checkAllFilesTransferred();
    // Returns either an EMBOSS input file ID if not all input files are transferred or return [rap] if all files are transferred
    QString embossInputFileID(const QString &filename);
    // Sends an emboss input filename (including the path on the client) followed by the md5 sum of the file to the server
    void writeEmbossInputFile(const QString &filename);
    // Sends the emboss input files to the server
    void updatePw();
    // The users password is expired, send a new password to the server

    void readServerStatus(const QString &data);
    // Receives and returns the server status
    void readInstalledEmbassyPackages(const QByteArray &data);
    // Receives the EMBASSY packages installed on the server
    void readEmbossOutput(const QByteArray &data);
    // Receive the messages, the emboss program produces on stdout and stderr
    void readEmbossOutputFileID(const QByteArray &data);
    // Receive EMBOSS output file identifier (filename (without path) followed by the md5 sum of the file)
    void readEmbossOutputFile(const QByteArray &data);
    // Receive the emboss output files

    QPointer<QSslSocket> tcpSocket;
    qint64 blockSize;
    QPointer<QNetworkSession> networkSession;
    quint16 fileListID;
    QString outputFilename;
    QByteArray outputMD5sum;
};

#endif // RUNEMBOSSAPPONSERVER_H
