/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2015 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/


#ifndef SETTINGS_H
#define SETTINGS_H

#include <QWidget>
#include <QString>
#include <QFont>
#include <QDir>
#include <QTextCharFormat>
#include <QCryptographicHash>

class kembossSettings : public QWidget
{
   Q_OBJECT

private:
    // EMBOSS and kemboss directories
    QString ckembossData;
    QString ckembossDefaultSettings;
    QString ckembossPortableConfigDir;
    QString cembossData;
    QString cembossBin;
    QString cembossHelp;
    QString cembossTest;

    bool cuseOnlineDoc;
    QString conlineDocSource;
    bool cportableApp;
    QString ccurrentDir;
    bool copenLastSettings;
    bool cautoCloseProgressWindow;
    bool cautoShowOutputTab;
    bool cautoDeleteExistingOutputFile;
    int cmaxExeTime;
    QFont ctextEditFont;
    bool cwrapLines;

    // Configuration of kemboss as a client
    bool cremoteEmboss;
    QString cHostname;
    quint16 cPort;
    QString cUsername;
    QString cUserGroup;
    QByteArray cPwHash;
    QString cEmail;

public:
    kembossSettings();
    // Class constructor
    virtual ~kembossSettings() {}
    // Class destructor
    void readSettings();
    // Read kemboss configuration from file in /etc (X11) or registry (Win)
    void writeSettings() const;
    // Write kemboss configuration from file in /etc (X11) or registry (Win)
    bool checkEmbossDirectories();
    // Check if all emboss directories exist. If not check the usual suspects.
    bool checkKembossDirectories();
    // Check if all kemboss directories exist. If not check the usual suspects.
    bool checkConfiguration();
    // Check configuration of kemboss client for consistency
    bool checkServer();
    // Check if connection to a remote server can be established and if the username and password are correct

    // get members
    QString kembossData() const {return ckembossData;}
    QString kembossDefaultSettings() const {return ckembossDefaultSettings;}
    QString kembossPortableConfigDir() const {return ckembossPortableConfigDir;}
    QString embossData() const {return cembossData;}
    QString embossBin() const {return cembossBin;}
    QString embossHelp() const {return cembossHelp;}
    QString embossTest() const {return cembossTest;}
    bool useOnlineDoc() const {return cuseOnlineDoc;}
    QString onlineDocSource() const {return conlineDocSource;}
    bool portableApp() const {return cportableApp;}
    QString currentDir() const {return ccurrentDir;}
    bool openLastSettings() const {return copenLastSettings;}
    bool autoCloseProgressWindow() const {return cautoCloseProgressWindow;}
    bool autoShowOutputTab() const {return cautoShowOutputTab;}
    bool autoDeleteExistingOutputFile() const {return cautoDeleteExistingOutputFile;}
    int maxExeTime() const {return cmaxExeTime;}
    QFont textEditFont() const {return ctextEditFont;}
    bool wrapLines() const {return cwrapLines;}
    bool remoteEmboss() const {return cremoteEmboss;}
    // Returns true if EMBOSS is installed on a remote server and kemboss is used as a client.
    QString hostname() const {return cHostname;}
    quint16 port() const {return cPort;}
    QString userName() const {return cUsername;}
    QString userGroup() const {return cUserGroup;}
    QByteArray userPwHash() const {return cPwHash;}
    QString userEmail() const {return cEmail;}

    // set members
    void setkembossData(const QString &dir) {ckembossData = QDir::toNativeSeparators(dir); if (!ckembossData.isEmpty() && !ckembossData.endsWith(QDir::separator())) ckembossData += QDir::separator(); }
    void setkembossDefaultSettings(const QString &dir) {ckembossDefaultSettings = QDir::toNativeSeparators(dir); if (!ckembossDefaultSettings.isEmpty() && !ckembossDefaultSettings.endsWith(QDir::separator())) ckembossDefaultSettings += QDir::separator(); }
    void setembossData(const QString &dir) {cembossData = QDir::toNativeSeparators(dir); if (!cembossData.isEmpty() && !cembossData.endsWith(QDir::separator())) cembossData += QDir::separator(); }
    void setembossBin(const QString &dir) {cembossBin = QDir::toNativeSeparators(dir); if (!cembossBin.isEmpty() && !cembossBin.endsWith(QDir::separator())) cembossBin += QDir::separator(); }
    void setembossHelp(const QString &dir) {cembossHelp = QDir::toNativeSeparators(dir); if (!cembossHelp.isEmpty() && !cembossHelp.endsWith(QDir::separator())) cembossHelp += QDir::separator(); }
    void setembossTest(const QString &dir) {cembossTest = QDir::toNativeSeparators(dir); if (!cembossTest.isEmpty() && !cembossTest.endsWith(QDir::separator())) cembossTest += QDir::separator(); }
    void setuseOnlineDoc(const bool checked) {cuseOnlineDoc = checked;}
    void setonlineDocSource(const QString &src) {conlineDocSource = src;}
    void setportableApp(const bool portable) {cportableApp = portable;}
    void setkembossPortableConfigDir(const QString &dir) {ckembossPortableConfigDir = QDir::toNativeSeparators(dir); }
    void setcurrentDir(const QString &dir) {ccurrentDir = QDir::toNativeSeparators(dir);}
    void setopenLastSettings(const bool checked) {copenLastSettings = checked;}
    void setautoCloseProgressWindow(const bool checked) {cautoCloseProgressWindow = checked;}
    void setautoShowOutputTab(const bool checked) {cautoShowOutputTab = checked;}
    void setautoDeleteExistingOutputFile(const bool checked) {cautoDeleteExistingOutputFile = checked;}
    void setmaxExeTime(const int &time) {cmaxExeTime = time;}
    void settextEditFont(const QString &family, int pointSize = -1, int weight = -1, bool italic = false) {QFont font(family, pointSize, weight, italic); ctextEditFont=font;}
    void setwrapLines(const bool checked) {cwrapLines = checked;}
    void setRemoteEmboss(const bool remote) {cremoteEmboss = remote;}
    void setHostname(const QString &hostname) {cHostname = hostname;}
    void setPort(const quint16 &port) {cPort = port;}
    void setUserName(const QString &name) {cUsername = name;}
    void setUserGroup(const QString &group) {cUserGroup = group;}
    void setUserPassword(const QString &pw) {cPwHash = QCryptographicHash::hash(pw.toLatin1(), QCryptographicHash::Sha1);}
    void setUserEmail(const QString &email) {cEmail = email;}

    QTextCharFormat seq_name;
    QTextCharFormat comment;
    QTextCharFormat cystein;
    QTextCharFormat basic_aa;
    QTextCharFormat dibasic_aa;
    QTextCharFormat acidic_aa;
    QTextCharFormat polar_aa;
    QTextCharFormat apolar_aa;
    QTextCharFormat a_nuc;
    QTextCharFormat t_nuc;
    QTextCharFormat g_nuc;
    QTextCharFormat c_nuc;
};

#endif
